/*

=head1 NAME

tinysh -- a tiny shell with a minimal set of features

=head1 SYNOPSIS

B<tinysh>

=head1 DESCRIPTION

B<tinysh> reads command lines from its standard input
and executes them. ``% '' is given as prompt.

Each command line consists of space-separated tokens. Special tokens
begin with ``<'', indicating the input file, ``>'', indicating the output
file, and ``>>'', specifying an output file that is to be extended. The
first non-special token specifies the command name which must be either
an absolute path of an executable file or locatable within the list of
directories provided by the environment variable B<PATH>.

=head1 DIAGNOSTICS

B<tinysh> prints the exit code of a terminated program if it is
non-zero. It tells that a program ``terminated abnormally'' if it
did not exit. An exit code of 255 is used by subprocesses if they
are unable to start the command for some reason.

=head1 BUGS

No signal handling, no support of backgrounded commands, no pipelines,
no builtins, no shell variables.

=head1 AUTHOR

Andreas F. Borchert

=cut

*/

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <unistd.h>
#include <afblib/inbuf.h>
#include <afblib/inbuf_sareadline.h>
#include <afblib/outbuf.h>
#include <afblib/outbuf_printf.h>
#include <afblib/strlist.h>
#include <afblib/tokenizer.h>

/* print error much like perror but by using an outbuf */
void print_error(const char* s) {
   outbuf err = {2};
   outbuf_printf(&err, "%s: %s\n", s, strerror(errno));
   outbuf_flush(&err);
   outbuf_free(&err);
}

/* assign an opened file with the given flags and mode to fd */
void fassign(int fd, char* path, int oflags, mode_t mode) {
   int newfd = open(path, oflags, mode);
   if (newfd < 0) {
      print_error(path); exit(255);
   }
   if (dup2(newfd, fd) < 0) {
      print_error("dup2"); exit(255);
   }
   close(newfd);
}

int main() {
   inbuf in = {0}; outbuf out = {1}; /* stdin and stdout */
   stralloc line = {0}; /* input line */
   strlist tokens = {0}; /* pointers to the invidual tokens of the input line */
   while (outbuf_printf(&out, "%% ") >= 0 && outbuf_flush(&out) &&
	 inbuf_sareadline(&in, &line)) {
      stralloc_0(&line); /* required by tokenizer() */
      if (!tokenizer(&line, &tokens)) break;
      if (tokens.len == 0) continue;

      pid_t child = fork();
      if (child == -1) {
         print_error("fork"); continue;
      }
      if (child == 0) {
         strlist argv = {0}; /* list of arguments */
         char* cmdname = 0; /* first argument */
         char* path; /* of output files */
         int oflags;

         for (int i = 0; i < tokens.len; ++i) {
            switch (tokens.list[i][0]) {
               case '<':
                  fassign(0, &tokens.list[i][1], O_RDONLY, 0);
                  break;
               case '>':
                  path = &tokens.list[i][1];
                  oflags = O_WRONLY|O_CREAT;
                  if (*path == '>') {
                     ++path; oflags |= O_APPEND;
                  } else {
                     oflags |= O_TRUNC;
                  }
                  fassign(1, path, oflags, 0666);
                  break;
               default:
                  strlist_push(&argv, tokens.list[i]);
                  if (cmdname == 0) cmdname = tokens.list[i];
            }
         }
         if (cmdname == 0) exit(0);
         strlist_push0(&argv);
         execvp(cmdname, argv.list);
	 print_error(cmdname);
         exit(255);
      }

      /* wait for termination of child */
      int stat;
      pid_t pid = wait(&stat);
      if (pid == child) {
         if (WIFEXITED(stat)) {
            int code = WEXITSTATUS(stat);
            if (code) {
               outbuf_printf(&out, "terminated with exit code %d\n", code);
            }
         } else {
            outbuf_printf(&out, "terminated abnormally\n");
         }
      } else {
         print_error("wait");
      }
   }
} // main
